section \<open>Matching\<close>

text \<open>
  Matching of first-order terms as example of using data types,
  non-trivial recursive functions, and code generation.
\<close>

theory Matching
imports
  Main
  "~~/src/HOL/Library/Multiset"
begin

subsection \<open>First-order terms and substitutions\<close>

text \<open>
  A data type of \<^bold>f\<^bold>i\<^bold>r\<^bold>s\<^bold>t\<^bold>-\<^bold>o\<^bold>r\<^bold>d\<^bold>e\<^bold>r \<^bold>t\<^bold>e\<^bold>r\<^bold>m\<^bold>s with function
  symbols of type @{typ 'f} and variables of
  type @{typ 'v}.
\<close>

datatype ('f, 'v) "term" =
  Var 'v
| Fun 'f "('f, 'v) term list"

text \<open>
  A \<^bold>s\<^bold>u\<^bold>b\<^bold>s\<^bold>t\<^bold>i\<^bold>t\<^bold>u\<^bold>t\<^bold>i\<^bold>o\<^bold>n is a partial function from variables to terms.
  (See theory @{theory Map} for more details on partial functions.)
\<close>

type_synonym ('f, 'v) subst = "'v \<Rightarrow> ('f, 'v) term option"


subsubsection \<open>Exercise: variables of a term\<close>

text \<open>
  \<triangleright> Define a recursive function @{term "vars :: ('f, 'v) term \<Rightarrow> 'v set"}
    by pattern matching on terms that computes the set of variables
    occurring in a term.

  Hint: Defined constants that might be useful are
  @{term "\<Union>(A::'a set set)"}, @{term set}, and @{term map}.
\<close>


text \<open>
  Applying a substitution to a variable does not change variables on which
  the substitution is not defined (i.e., variables @{term x} such that
  @{term "\<sigma> x = None"}).
\<close>

abbreviation subst_var :: "('f, 'v) subst \<Rightarrow> 'v \<Rightarrow> ('f, 'v) term"
where
  "subst_var \<sigma> x \<equiv> (case \<sigma> x of Some t \<Rightarrow> t | None \<Rightarrow> Var x)"

lemma subst_var_iff [simp]: "x \<in> dom \<sigma> \<and> subst_var \<sigma> x = t \<longleftrightarrow> \<sigma> x = Some t"
  by auto


subsubsection \<open>Exercise: applying substitutions to terms\<close>

text \<open>
  \<triangleright> Define a function
    @{term "apply_subst :: ('f, 'v) term \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) term"}
    by pattern matching on type @{typ "('f, 'v) term"} such that
    @{term "apply_subst t \<sigma>"} applies the substitution @{term \<sigma>}
    to the term @{term t}.

  \<triangleright> Introduce infix syntax using @{text "\<cdot>"} for @{term apply_subst}.

  Hint: Use @{term subst_var} for the base case.
\<close>


subsection \<open>Implementation -- a concrete matching algorithm\<close>

text \<open>An \<^bold>e\<^bold>q\<^bold>u\<^bold>a\<^bold>t\<^bold>i\<^bold>o\<^bold>n is a pair of terms.\<close>

type_synonym ('f, 'v) eq = "('f, 'v) term \<times> ('f, 'v) term"

function match :: "('f, 'v) eq list \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst option"
where
  "match [] \<sigma> = Some \<sigma>"
| "match ((Var x, t) # eqs) \<sigma> =
    (if \<sigma> x = None \<or> \<sigma> x = Some t then match eqs (\<sigma> (x \<mapsto> t))
    else None)"
| "match ((Fun f ss, Fun g ts) # eqs) \<sigma> =
    (if g = f \<and> length ts = length ss then match (zip ss ts @ eqs) \<sigma>
    else None)"
| "match ((Fun f ss, Var x) # eqs) \<sigma> = None"
  by pat_completeness auto


subsubsection \<open>Exercise: Termination of @{term match}\<close>

text \<open>
  \<triangleright> Find a measure function (i.e., a function into the natural numbers)
    on lists of equations that guarantees termination of @{term match}.

  Hint: Functions that might be useful: @{term listsum}.
\<close>


subsection \<open>Code generation and some peephole optimization\<close>

subsubsection \<open>Exercise: Optimization\<close>

text \<open>
  After inspecting the code equations of @{term match} you might notice
  possible optimization(s).

  \<triangleright> Find at least one optimization of @{term match} and prove a
    corresponding code equation.
\<close>


subsection \<open>Specification -- a small theory of matchers\<close>

text \<open>
  A \<^bold>m\<^bold>a\<^bold>t\<^bold>c\<^bold>h\<^bold>i\<^bold>n\<^bold>g \<^bold>p\<^bold>r\<^bold>o\<^bold>b\<^bold>l\<^bold>e\<^bold>m is a set of equations.
  Given a matching problem, the \<^bold>s\<^bold>e\<^bold>t \<^bold>o\<^bold>f \<^bold>m\<^bold>a\<^bold>t\<^bold>c\<^bold>h\<^bold>e\<^bold>r\<^bold>s comprises exactly those
  substitutions that are defined on all variables occurring in left-hand
  sides of the equations in @{term P} and furthermore make the left-hand
  sides equal to the corresponding right-hand sides.

definition matchers :: "('f, 'v) eq set \<Rightarrow> ('f, 'v) subst set"
where
  "matchers P = {\<sigma>. \<forall>(s, t) \<in> P. vars s \<subseteq> dom \<sigma> \<and> s \<cdot> \<sigma> = t}"
\<close>


subsubsection \<open>Exercise: Basic properties of matchers\<close>

text \<open>
  \<triangleright> Prove the following properties of @{term matchers}:
    @{text "matchers {} = UNIV"},
    @{text "matchers (insert (s, t) P) =
      {\<sigma>. vars s \<subseteq> dom \<sigma> \<and> s \<cdot> \<sigma> = t} \<inter> matchers P"}, and
    @{text "matchers (P \<union> Q) = matchers P \<inter> matchers Q"}.

  \<triangleright> Declare the resulting facts to be standard simp rules.
\<close>


subsection \<open>An abstract matching algorithm\<close>

text \<open>
  An abstract 1-step matching relation, whose reflexive and transitive
  closure should cover actual implementations.
\<close>

inductive_set MATCH1
where
  "\<sigma> x = None \<or> \<sigma> x = Some t \<Longrightarrow>
    ((P + {#(Var x, t)#}, \<sigma>),
     (P, \<sigma> (x \<mapsto> t))) \<in> MATCH1"
| "length ts = length ss \<Longrightarrow>
    ((P + {#(Fun f ss, Fun f ts)#}, \<sigma>),
     (P + multiset_of (zip ss ts), \<sigma>)) \<in> MATCH1"

abbreviation "MATCH \<equiv> MATCH1\<^sup>*"


subsubsection \<open>Exercise: An invariant of @{term MATCH}\<close>

text \<open>
  Considering the bindings of a substitution @{term \<sigma>} as set of equations,
  we can examine the set of corresponding matchers. Intuitively, these are
  those substitutions that extend @{term \<sigma>} by additional bindings. This
  relationship is expressed by the relation @{term "op \<subseteq>\<^sub>m"}.

  \<triangleright> Show that the following set is invariant under transitions w.r.t.
    @{term MATCH}: @{term "matchers (set_of P) \<inter> {\<upsilon>. \<sigma> \<subseteq>\<^sub>m \<upsilon>}"}.
    Where @{term P} is the matching problem and @{term \<sigma>} the substitution
    of the current "state".
\<close>


subsubsection \<open>Exercise: Correctness of @{term MATCH}\<close>

text \<open>
  \<triangleright> Use the previous results to conclude that whenever we reach the empty
    matching problem @{term "({#}, \<sigma>)"} starting from the pair
    @{term "(P, Map.empty)"} and using transitions w.r.t. @{term MATCH},
    then @{term \<sigma>} is a matcher of the equations in @{term P}.
\<close>


subsubsection \<open>Exercise: Soundness of @{term match}\<close>

text \<open>
  \<triangleright> Show that @{term match} is sound w.r.t. to the abstract specification
    @{term MATCH}, i.e.,
    @{prop "match eqs \<sigma> = Some \<tau> \<Longrightarrow>
      ((multiset_of eqs, \<sigma>), ({#}, \<tau>)) \<in> MATCH"}.

  \<triangleright> Conclude that the result of @{term "match eqs Map.empty"} is indeed a
    matcher of the equations in @{term eqs}.
\<close>

end
