section \<open>Matching\<close>

text \<open>
  Matching of first-order terms as example of using data types,
  non-trivial recursive functions, and code generation.
\<close>

theory Matching_Solution
imports
  Main
  "~~/src/HOL/Library/Multiset"
begin

subsection \<open>First-order terms and substitutions\<close>

text \<open>
  A data type of \<^bold>f\<^bold>i\<^bold>r\<^bold>s\<^bold>t\<^bold>-\<^bold>o\<^bold>r\<^bold>d\<^bold>e\<^bold>r \<^bold>t\<^bold>e\<^bold>r\<^bold>m\<^bold>s with function
  symbols of type @{typ 'f} and variables of
  type @{typ 'v}.
\<close>

datatype ('f, 'v) "term" =
  Var 'v
| Fun 'f "('f, 'v) term list"

text \<open>
  A \<^bold>s\<^bold>u\<^bold>b\<^bold>s\<^bold>t\<^bold>i\<^bold>t\<^bold>u\<^bold>t\<^bold>i\<^bold>o\<^bold>n is a partial function from variables to terms.
  (See theory @{theory Map} for more details on partial functions.)
\<close>

type_synonym ('f, 'v) subst = "'v \<Rightarrow> ('f, 'v) term option"


subsubsection \<open>Exercise: variables of a term\<close>

text \<open>
  \<triangleright> Define a recursive function @{term "vars :: ('f, 'v) term \<Rightarrow> 'v set"}
    by pattern matching on terms that computes the set of variables
    occurring in a term.

  Hint: Defined constants that might be useful are
  @{term "\<Union>(A::'a set set)"}, @{term set}, and @{term map}.
\<close>


subsubsection \<open>Solution\<close>

fun vars :: "('f, 'v) term \<Rightarrow> 'v set"
where
  "vars (Var x) = {x}"
| "vars (Fun f ts) = \<Union>(set (map vars ts))"

text \<open>
  Applying a substitution to a variable does not change variables on which
  the substitution is not defined (i.e., variables @{term x} such that
  @{term "\<sigma> x = None"}).
\<close>

abbreviation subst_var :: "('f, 'v) subst \<Rightarrow> 'v \<Rightarrow> ('f, 'v) term"
where
  "subst_var \<sigma> x \<equiv> (case \<sigma> x of Some t \<Rightarrow> t | None \<Rightarrow> Var x)"

lemma subst_var_iff [simp]: "x \<in> dom \<sigma> \<and> subst_var \<sigma> x = t \<longleftrightarrow> \<sigma> x = Some t"
  by auto


subsubsection \<open>Exercise: applying substitutions to terms\<close>

text \<open>
  \<triangleright> Define a function
    @{term "apply_subst :: ('f, 'v) term \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) term"}
    by pattern matching on type @{typ "('f, 'v) term"} such that
    @{term "apply_subst t \<sigma>"} applies the substitution @{term \<sigma>}
    to the term @{term t}.

  \<triangleright> Introduce infix syntax using @{text "\<cdot>"} for @{term apply_subst}.

  Hint: Use @{term subst_var} for the base case.
\<close>


subsubsection \<open>Solution\<close>

fun apply_subst :: "('f, 'v) term \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) term"
  (infixl "\<cdot>" 65)
where
  "Var x \<cdot> \<sigma> = subst_var \<sigma> x"
| "Fun f ts \<cdot> \<sigma> = Fun f (map (\<lambda>t. t \<cdot> \<sigma>) ts)"


subsection \<open>Implementation -- a concrete matching algorithm\<close>

text \<open>An \<^bold>e\<^bold>q\<^bold>u\<^bold>a\<^bold>t\<^bold>i\<^bold>o\<^bold>n is a pair of terms.\<close>

type_synonym ('f, 'v) eq = "('f, 'v) term \<times> ('f, 'v) term"

function match :: "('f, 'v) eq list \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst option"
where
  "match [] \<sigma> = Some \<sigma>"
| "match ((Var x, t) # eqs) \<sigma> =
    (if \<sigma> x = None \<or> \<sigma> x = Some t then match eqs (\<sigma> (x \<mapsto> t))
    else None)"
| "match ((Fun f ss, Fun g ts) # eqs) \<sigma> =
    (if g = f \<and> length ts = length ss then match (zip ss ts @ eqs) \<sigma>
    else None)"
| "match ((Fun f ss, Var x) # eqs) \<sigma> = None"
  by pat_completeness auto


subsubsection \<open>Exercise: Termination of @{term match}\<close>

text \<open>
  \<triangleright> Find a measure function (i.e., a function into the natural numbers)
    on lists of equations that guarantees termination of @{term match}.

  Hint: Functions that might be useful: @{term listsum}.
\<close>


subsubsection \<open>Solution\<close>

fun size_term :: "('f, 'v) term \<Rightarrow> nat"
where
  "size_term (Var x) = 1"
| "size_term (Fun f ts) = Suc (listsum (map size_term ts))"

fun size_eqs :: "('f, 'v) eq list \<Rightarrow> nat"
where
  "size_eqs [] = 0"
| "size_eqs ((s, t) # eqs) = size_term s + size_term t + size_eqs eqs"

lemma size_eqs_append: "size_eqs (xs @ ys) = size_eqs xs + size_eqs ys"
  by (induct xs) auto

lemma size_eqs_zip:
  assumes "length ts = length ss"
  shows "size_eqs (zip ss ts) = listsum (map size_term ss) + listsum (map size_term ts)"
  using assms by (induct ts ss rule: list_induct2) simp_all

termination match
  by (relation "measure (size_eqs \<circ> fst)")
     (auto simp: size_eqs_append size_eqs_zip)


subsection \<open>Code generation and some peephole optimization\<close>

export_code match in Haskell


subsubsection \<open>Exercise: Optimization\<close>

text \<open>
  After inspecting the code equations of @{term match} you might notice
  possible optimization(s).

  \<triangleright> Find at least one optimization of @{term match} and prove a
    corresponding code equation.
\<close>


subsection \<open>Solution\<close>

lemma match_Var_code [code]:
  "match ((Var x, t) # eqs) \<sigma> =
    (if \<sigma> x = Some t then match eqs \<sigma>
     else if \<sigma> x = None then match eqs (\<sigma> (x \<mapsto> t))
     else None)"
  by (simp add: fun_upd_idem)

export_code match in Haskell


subsection \<open>Specification -- a small theory of matchers\<close>

text \<open>
  A \<^bold>m\<^bold>a\<^bold>t\<^bold>c\<^bold>h\<^bold>i\<^bold>n\<^bold>g \<^bold>p\<^bold>r\<^bold>o\<^bold>b\<^bold>l\<^bold>e\<^bold>m is a set of equations.
  Given a matching problem, the \<^bold>s\<^bold>e\<^bold>t \<^bold>o\<^bold>f \<^bold>m\<^bold>a\<^bold>t\<^bold>c\<^bold>h\<^bold>e\<^bold>r\<^bold>s comprises exactly those
  substitutions that are defined on all variables occurring in left-hand
  sides of the equations in @{term P} and furthermore make the left-hand
  sides equal to the corresponding right-hand sides.
\<close>

definition matchers :: "('f, 'v) eq set \<Rightarrow> ('f, 'v) subst set"
where
  "matchers P = {\<sigma>. \<forall>(s, t) \<in> P. vars s \<subseteq> dom \<sigma> \<and> s \<cdot> \<sigma> = t}"


subsubsection \<open>Exercise: Basic properties of matchers\<close>

text \<open>
  \<triangleright> Prove the following properties of @{term matchers}:
    @{prop "matchers {} = UNIV"},
    @{prop "matchers (insert (s, t) P) =
      {\<sigma>. vars s \<subseteq> dom \<sigma> \<and> s \<cdot> \<sigma> = t} \<inter> matchers P"}, and
    @{prop "matchers (P \<union> Q) = matchers P \<inter> matchers Q"}.

  \<triangleright> Declare the resulting facts to be standard simp rules.
\<close>


subsubsection \<open>Solution\<close>

lemma matchers_empty [simp]: "matchers {} = UNIV"
  by (simp add: matchers_def)

lemma matchers_insert [simp]:
  "matchers (insert (s, t) P) =
    {\<sigma>. vars s \<subseteq> dom \<sigma> \<and> s \<cdot> \<sigma> = t} \<inter> matchers P"
  by (auto simp: matchers_def)

lemma matchers_union [simp]: "matchers (P \<union> Q) = matchers P \<inter> matchers Q"
  by (force simp: matchers_def)


subsection \<open>An abstract matching algorithm\<close>

text \<open>
  An abstract 1-step matching relation, whose reflexive and transitive
  closure should cover actual implementations.
\<close>

inductive_set MATCH1
where
  "\<sigma> x = None \<or> \<sigma> x = Some t \<Longrightarrow>
    ((P + {#(Var x, t)#}, \<sigma>),
     (P, \<sigma> (x \<mapsto> t))) \<in> MATCH1"
| "length ts = length ss \<Longrightarrow>
    ((P + {#(Fun f ss, Fun f ts)#}, \<sigma>),
     (P + multiset_of (zip ss ts), \<sigma>)) \<in> MATCH1"

abbreviation "MATCH \<equiv> MATCH1\<^sup>*"


subsubsection \<open>Exercise: An invariant of @{term MATCH}\<close>

text \<open>
  Considering the bindings of a substitution @{term \<sigma>} as set of equations,
  we can examine the set of corresponding matchers. Intuitively, these are
  those substitutions that extend @{term \<sigma>} by additional bindings. This
  relationship is expressed by the relation @{term "op \<subseteq>\<^sub>m"}.

  \<triangleright> Show that the following set is invariant under transitions w.r.t.
    @{term MATCH}: @{term "matchers (set_of P) \<inter> {\<upsilon>. \<sigma> \<subseteq>\<^sub>m \<upsilon>}"}.
    Where @{term P} is the matching problem and @{term \<sigma>} the substitution
    of the current "state".
\<close>


subsubsection \<open>Solution\<close>

lemma upd_map_le_conv [simp]:
  assumes "\<sigma> x = None \<or> \<sigma> x = Some t"
  shows "{\<tau>. \<sigma> (x \<mapsto> t) \<subseteq>\<^sub>m \<tau>} = {\<tau>. \<sigma> \<subseteq>\<^sub>m \<tau>} \<inter> {\<tau>. \<tau> x = Some t}"
  using assms by (auto simp: map_le_def dom_def)

lemma matchers_set_zip [simp]:
  assumes "length ss = length ts"
  shows "matchers (set (zip ss ts)) =
    {\<sigma>. \<Union>(set (map vars ss)) \<subseteq> dom \<sigma> \<and> map (\<lambda>t. t \<cdot> \<sigma>) ss = ts}"
  using assms by (induct ss ts rule: list_induct2) auto

lemma MATCH1:
  assumes "((P, \<sigma>), (Q, \<tau>)) \<in> MATCH1"
  shows "matchers (set_of P) \<inter> {\<upsilon>. \<sigma> \<subseteq>\<^sub>m \<upsilon>} =
    matchers (set_of Q) \<inter> {\<upsilon>. \<tau> \<subseteq>\<^sub>m \<upsilon>}"
  using assms by (induct) (simp_all del: fun_upd_apply add: ac_simps)

lemma MATCH:
  assumes "(x, y) \<in> MATCH"
  shows "matchers (set_of (fst x)) \<inter> {\<upsilon>. snd x \<subseteq>\<^sub>m \<upsilon>} =
    matchers (set_of (fst y)) \<inter> {\<upsilon>. snd y \<subseteq>\<^sub>m \<upsilon>}"
  using assms by (induct) (simp_all add: MATCH1)


subsubsection \<open>Exercise: Correctness of @{term MATCH}\<close>

text \<open>
  \<triangleright> Use the previous results to conclude that whenever we reach the empty
    matching problem @{term "({#}, \<sigma>)"} starting from the pair
    @{term "(P, Map.empty)"} and using transitions w.r.t. @{term MATCH},
    then @{term \<sigma>} is a matcher of the equations in @{term P}.
\<close>


subsubsection \<open>Solution\<close>

lemma MATCH_matchers:
  assumes "((P, Map.empty), ({#}, \<sigma>)) \<in> MATCH"
  shows "\<sigma> \<in> matchers (set_of P)"
  using MATCH [OF assms] by simp


subsubsection \<open>Exercise: Soundness of @{term match}\<close>

text \<open>
  \<triangleright> Show that @{term match} is sound w.r.t. to the abstract specification
    @{term MATCH}, i.e.,
    @{prop "match eqs \<sigma> = Some \<tau> \<Longrightarrow>
      ((multiset_of eqs, \<sigma>), ({#}, \<tau>)) \<in> MATCH"}.

  \<triangleright> Conclude that the result of @{term "match eqs Map.empty"} is indeed a
    matcher of the equations in @{term eqs}.
\<close>


subsubsection \<open>Solution\<close>

text \<open>The concrete algorithm is sound w.r.t. the abstract specification.\<close>
lemma match_Some_imp_MATCH:
  assumes "match eqs \<sigma> = Some \<tau>"
  shows "((multiset_of eqs, \<sigma>), ({#}, \<tau>)) \<in> MATCH"
  using assms
proof (induct eqs \<sigma> rule: match.induct)
  case (2 x t eqs \<sigma>)
  then have *: "\<sigma> x = None \<or> \<sigma> x = Some t"
    and "match eqs (\<sigma>(x \<mapsto> t)) = Some \<tau>"
    by (auto split: if_splits)
  then have "((multiset_of eqs, \<sigma>(x \<mapsto> t)), {#}, \<tau>) \<in> MATCH"
    by fact -- \<open>implicit discharge of assumptions of IH\<close>
  moreover
  have "((multiset_of eqs + {#(Var x, t)#}, \<sigma>),
         (multiset_of eqs, \<sigma> (x \<mapsto> t))) \<in> MATCH1"
    using * by (auto intro: MATCH1.intros)
  ultimately
  show ?case by auto
next
  case (3 f ss g ts eqs \<sigma>)
  then have *: "g = f \<and> length ts = length ss"
    and "match (zip ss ts @ eqs) \<sigma> = Some \<tau>"
    by (auto split: if_splits)
  then have "((multiset_of (zip ss ts @ eqs), \<sigma>), {#}, \<tau>) \<in> MATCH"
    by fact
  moreover
  have "((multiset_of ((Fun f ss, Fun g ts) # eqs), \<sigma>),
         (multiset_of (zip ss ts @ eqs), \<sigma>)) \<in> MATCH1"
    using * by (auto simp: ac_simps intro: MATCH1.intros)
  ultimately
  show ?case by auto
qed simp_all

lemma match_Some_imp_matchers:
  assumes "match eqs Map.empty = Some \<tau>"
  shows "\<tau> \<in> matchers (set eqs)"
  using assms by (force dest: match_Some_imp_MATCH MATCH_matchers)


subsection \<open>Correctness of the @{term None} case\<close>

text \<open>
  Immediate causes of non-matching:

  \<bullet> function symbols clash

  \<bullet> function symbol on the left clashes with variable on the right

  \<bullet> number of arguments clash

  \<bullet> clashing bindings for variable
\<close>

inductive clash
where
  Fun1: "f \<noteq> g \<Longrightarrow> (Fun f ss, Fun g ts) \<in># P \<Longrightarrow> clash (P, \<sigma>)"
| Fun2: "(Fun f ss, Var x) \<in># P \<Longrightarrow> clash (P, \<sigma>)"
| length: "length ss \<noteq> length ts \<Longrightarrow>
    (Fun f ss, Fun g ts) \<in># P \<Longrightarrow> clash (P, \<sigma>)"
| Var1: "t \<noteq> t' \<Longrightarrow> (Var x, t) \<in># P \<Longrightarrow> (Var x, t') \<in># P \<Longrightarrow> clash (P, \<sigma>)"
| Var2: "t \<noteq> t' \<Longrightarrow> \<sigma> x = Some t \<Longrightarrow> (Var x, t') \<in># P \<Longrightarrow> clash (P, \<sigma>)"

lemma matchers_empty_Fun_simps [simp]:
  "f \<noteq> g \<Longrightarrow> (Fun f ss, Fun g ts) \<in> S \<Longrightarrow> matchers S = {}"
  "length ss \<noteq> length ts \<Longrightarrow> (Fun f ss, Fun g ts) \<in> S \<Longrightarrow> matchers S = {}"
  "(Fun f ss, Var x) \<in> S \<Longrightarrow> matchers S = {}"
  using assms by (force simp: matchers_def)+

lemma matchers_empty_Var_simp [simp]:
  assumes "t \<noteq> t'"
    and "(Var x, t) \<in> S"
    and "(Var x, t') \<in> S"
  shows "matchers S = {}"
proof -
  {
    fix \<sigma>
    have "\<exists>e \<in> S. fst e \<cdot> \<sigma> \<noteq> snd e"
      using assms by (cases "Var x \<cdot> \<sigma> = t"; cases "Var x \<cdot> \<sigma> = t'") force+
  }
  then show ?thesis by (force simp: matchers_def)
qed

lemma matchers_Int_subst_empty_simp [simp]:
  assumes "t \<noteq> t'"
    and "\<sigma> x = Some t"
    and "(Var x, t') \<in> S"
  shows "matchers S \<inter> {\<tau>. \<sigma> \<subseteq>\<^sub>m \<tau>} = {}"
  using assms by (auto simp: matchers_def map_le_def dom_def)

lemma clash_imp_matchers_empy:
  assumes "clash (P, \<sigma>)"
  shows "matchers (set_of P) \<inter> {\<tau>. \<sigma> \<subseteq>\<^sub>m \<tau>} = {}"
  using assms by (cases) simp_all

lemma match_None_imp_clash:
  assumes "match eqs \<sigma> = None"
  shows "\<exists>T \<tau>. ((multiset_of eqs, \<sigma>), (T, \<tau>)) \<in> MATCH \<and> clash (T, \<tau>)"
  using assms
proof (induction eqs \<sigma> rule: match.induct)
  case (2 x t eqs \<sigma>)
  then show ?case
    using MATCH1.intros(1) [of \<sigma> x t "multiset_of eqs"]
      and clash.Var2 [of _ t \<sigma> x]
      by (auto simp del: fun_upd_apply split: if_splits
          intro: converse_rtrancl_into_rtrancl) force
next
  case (3 f ss g ts eqs \<sigma>)
  then show ?case
    using MATCH1.intros(2) [of ts ss "multiset_of eqs" f \<sigma>]
      and clash.Fun1 [of f g _ ss ts \<sigma>]
      and clash.length [of ss ts _ f g \<sigma>]
      by (auto simp: ac_simps split: if_splits
          intro: converse_rtrancl_into_rtrancl) force+
next
  case (4 f ss x eqs \<sigma>)
  then show ?case
    using clash.Fun2 [of _ f ss x \<sigma>] by force
qed simp

end
